@@ -6,7 +6,6 @@ import logging
 import uuid
 import warnings
 from contextlib import suppress
-from inspect import isawaitable
 from typing import Any
 
 from packaging.version import parse as parse_version
@@ -19,6 +18,7 @@ from dask.widgets import get_template
 from distributed.compatibility import PeriodicCallback
 from distributed.core import Status
 from distributed.deploy.adaptive import Adaptive
+from distributed.metrics import time
 from distributed.objects import SchedulerInfo
 from distributed.utils import (
     Log,
@@ -33,6 +33,9 @@ from distributed.utils import (
 logger = logging.getLogger(__name__)
 
 
+no_default = "__no_default__"
+
+
 class Cluster(SyncMethodMixin):
     """Superclass for cluster objects
 
@@ -68,6 +71,7 @@ class Cluster(SyncMethodMixin):
         scheduler_sync_interval=1,
     ):
         self._loop_runner = LoopRunner(loop=loop, asynchronous=asynchronous)
+        self.__asynchronous = asynchronous
 
         self.scheduler_info = {"workers": {}}
         self.periodic_callbacks = {}
@@ -111,6 +115,15 @@ class Cluster(SyncMethodMixin):
             raise ValueError("expected an IOLoop, got None")
         self.__loop = value
 
+    @property
+    def called_from_running_loop(self):
+        try:
+            return (
+                getattr(self.loop, "asyncio_loop", None) is asyncio.get_running_loop()
+            )
+        except RuntimeError:
+            return self.__asynchronous
+
     @property
     def name(self):
         return self._cluster_info["name"]
@@ -157,11 +170,6 @@ class Cluster(SyncMethodMixin):
                     value=self._cluster_info.copy(),
                 )
                 err_count = 0
-            except asyncio.CancelledError:
-                # Task is being closed. When we drop Python < 3.8 we can drop
-                # this check (since CancelledError is not a subclass of
-                # Exception then).
-                break
             except Exception:
                 err_count += 1
                 # Only warn if multiple subsequent attempts fail, and only once
@@ -175,7 +183,9 @@ class Cluster(SyncMethodMixin):
                         exc_info=True,
                     )
             # Sleep, with error backoff
-            interval = min(max_interval, self._sync_interval * 1.5**err_count)
+            interval = _exponential_backoff(
+                err_count, self._sync_interval, 1.5, max_interval
+            )
             await asyncio.sleep(interval)
 
     async def _close(self):
@@ -205,16 +215,17 @@ class Cluster(SyncMethodMixin):
 
         self.status = Status.closed
 
-    def close(self, timeout=None):
+    def close(self, timeout: float | None = None) -> Any:
         # If the cluster is already closed, we're already done
         if self.status == Status.closed:
             if self.asynchronous:
                 return NoOpAwaitable()
-            else:
-                return
+            return None
 
-        with suppress(RuntimeError):  # loop closed during process shutdown
+        try:
             return self.sync(self._close, callback_timeout=timeout)
+        except RuntimeError:  # loop closed during process shutdown
+            return None
 
     def __del__(self, _warn=warnings.warn):
         if getattr(self, "status", Status.closed) != Status.closed:
@@ -479,7 +490,6 @@ class Cluster(SyncMethodMixin):
         return tab
 
     def _repr_html_(self, cluster_status=None):
-
         try:
             scheduler_info_repr = self.scheduler_info._repr_html_()
         except AttributeError:
@@ -519,10 +529,16 @@ class Cluster(SyncMethodMixin):
             display(mimebundle, raw=True)
 
     def __enter__(self):
+        if self.asynchronous:
+            raise TypeError(
+                "Used 'with' with asynchronous class; please use 'async with'"
+            )
+
         return self.sync(self.__aenter__)
 
     def __exit__(self, exc_type, exc_value, traceback):
-        return self.sync(self.__aexit__, exc_type, exc_value, traceback)
+        aw = self.close()
+        assert aw is None, aw
 
     def __await__(self):
         return self
@@ -533,9 +549,7 @@ class Cluster(SyncMethodMixin):
         return self
 
     async def __aexit__(self, exc_type, exc_value, traceback):
-        f = self.close()
-        if isawaitable(f):
-            await f
+        await self.close()
 
     @property
     def scheduler_address(self) -> str:
@@ -580,3 +594,66 @@ class Cluster(SyncMethodMixin):
 
     def __hash__(self):
         return id(self)
+
+    async def _wait_for_workers(self, n_workers=0, timeout=None):
+        self.scheduler_info = SchedulerInfo(await self.scheduler_comm.identity())
+        if timeout:
+            deadline = time() + parse_timedelta(timeout)
+        else:
+            deadline = None
+
+        def running_workers(info):
+            return len(
+                [
+                    ws
+                    for ws in info["workers"].values()
+                    if ws["status"] == Status.running.name
+                ]
+            )
+
+        while n_workers and running_workers(self.scheduler_info) < n_workers:
+            if deadline and time() > deadline:
+                raise TimeoutError(
+                    "Only %d/%d workers arrived after %s"
+                    % (running_workers(self.scheduler_info), n_workers, timeout)
+                )
+            await asyncio.sleep(0.1)
+
+            self.scheduler_info = SchedulerInfo(await self.scheduler_comm.identity())
+
+    def wait_for_workers(
+        self, n_workers: int | str = no_default, timeout: float | None = None
+    ) -> None:
+        """Blocking call to wait for n workers before continuing
+
+        Parameters
+        ----------
+        n_workers : int
+            The number of workers
+        timeout : number, optional
+            Time in seconds after which to raise a
+            ``dask.distributed.TimeoutError``
+        """
+        if n_workers is no_default:
+            warnings.warn(
+                "Please specify the `n_workers` argument when using `Client.wait_for_workers`. Not specifying `n_workers` will no longer be supported in future versions.",
+                FutureWarning,
+            )
+            n_workers = 0
+        elif not isinstance(n_workers, int) or n_workers < 1:
+            raise ValueError(
+                f"`n_workers` must be a positive integer. Instead got {n_workers}."
+            )
+        return self.sync(self._wait_for_workers, n_workers, timeout=timeout)
+
+
+def _exponential_backoff(
+    attempt: int, multiplier: float, exponential_base: float, max_interval: float
+) -> float:
+    """Calculate the duration of an exponential backoff"""
+    try:
+        interval = multiplier * exponential_base**attempt
+    except OverflowError:
+        return max_interval
+
+    return min(max_interval, interval)
