@@ -21,7 +21,13 @@ from distributed.utils import log_errors, recursive_to_dict
 
 if TYPE_CHECKING:
     # Recursive imports
-    from distributed.scheduler import Scheduler, SchedulerState, TaskState, WorkerState
+    from distributed.scheduler import (
+        Scheduler,
+        SchedulerState,
+        TaskState,
+        TaskStateState,
+        WorkerState,
+    )
 
 # Stealing requires multiple network bounces and if successful also task
 # submission which may include code serialization. Therefore, be very
@@ -155,7 +161,7 @@ class WorkStealing(SchedulerPlugin):
     def add_worker(self, scheduler: Any = None, worker: Any = None) -> None:
         self.stealable[worker] = tuple(set() for _ in range(15))
 
-    def remove_worker(self, scheduler: Scheduler, worker: str) -> None:
+    def remove_worker(self, scheduler: Scheduler, worker: str, **kwargs: Any) -> None:
         del self.stealable[worker]
 
     def teardown(self) -> None:
@@ -167,10 +173,8 @@ class WorkStealing(SchedulerPlugin):
     def transition(
         self,
         key: str,
-        start: str,
-        finish: str,
-        compute_start: Any = None,
-        compute_stop: Any = None,
+        start: TaskStateState,
+        finish: TaskStateState,
         *args: Any,
         **kwargs: Any,
     ) -> None:
@@ -428,7 +432,6 @@ class WorkStealing(SchedulerPlugin):
                 if not potential_thieves:
                     break
                 for victim in list(potential_victims):
-
                     stealable = self.stealable[victim.address][level]
                     if not stealable or not potential_thieves:
                         continue
@@ -439,15 +442,14 @@ class WorkStealing(SchedulerPlugin):
                         if (
                             ts not in self.key_stealable
                             or ts.processing_on is not victim
+                            or ts not in victim.processing
                         ):
+                            # FIXME: Instead of discarding here, clean up stealable properly
                             stealable.discard(ts)
                             continue
                         i += 1
                         if not (thief := _get_thief(s, ts, potential_thieves)):
                             continue
-                        if ts not in victim.processing:
-                            stealable.discard(ts)
-                            continue
 
                         occ_thief = self._combined_occupancy(thief)
                         occ_victim = self._combined_occupancy(victim)
@@ -483,6 +485,9 @@ class WorkStealing(SchedulerPlugin):
                                 thief, occ_thief, nproc_thief
                             ):
                                 potential_thieves.discard(thief)
+                            # FIXME: move_task_request already implements some logic
+                            # for removing ts from stealable. If we made sure to
+                            # properly clean up, we would not need this
                             stealable.discard(ts)
                     self.scheduler.check_idle_saturated(
                         victim, occ=self._combined_occupancy(victim)
