@@ -4,8 +4,10 @@ import asyncio
 import logging
 import random
 from collections import defaultdict
+from collections.abc import Callable, Collection, Coroutine, Mapping
 from functools import partial
 from itertools import cycle
+from typing import Any, TypeVar
 
 from tlz import concat, drop, groupby, merge
 
@@ -13,13 +15,19 @@ import dask.config
 from dask.optimization import SubgraphCallable
 from dask.utils import is_namedtuple_instance, parse_timedelta, stringify
 
-from distributed.core import rpc
+from distributed.core import ConnectionPool, rpc
 from distributed.utils import All
 
 logger = logging.getLogger(__name__)
 
 
-async def gather_from_workers(who_has, rpc, close=True, serializers=None, who=None):
+async def gather_from_workers(
+    who_has: Mapping[str, Collection[str]],
+    rpc: ConnectionPool,
+    close: bool = True,
+    serializers: list[str] | None = None,
+    who: str | None = None,
+) -> tuple[dict[str, object], dict[str, list[str]], list[str]]:
     """Gather data directly from peers
 
     Parameters
@@ -37,18 +45,18 @@ async def gather_from_workers(who_has, rpc, close=True, serializers=None, who=No
     """
     from distributed.worker import get_data_from_worker
 
-    bad_addresses = set()
+    bad_addresses: set[str] = set()
     missing_workers = set()
     original_who_has = who_has
-    who_has = {k: set(v) for k, v in who_has.items()}
-    results = dict()
-    all_bad_keys = set()
+    new_who_has = {k: set(v) for k, v in who_has.items()}
+    results: dict[str, object] = {}
+    all_bad_keys: set[str] = set()
 
     while len(results) + len(all_bad_keys) < len(who_has):
         d = defaultdict(list)
         rev = dict()
         bad_keys = set()
-        for key, addresses in who_has.items():
+        for key, addresses in new_who_has.items():
             if key in results:
                 continue
             try:
@@ -61,19 +69,23 @@ async def gather_from_workers(who_has, rpc, close=True, serializers=None, who=No
             all_bad_keys |= bad_keys
         coroutines = {
             address: asyncio.create_task(
-                get_data_from_worker(
-                    rpc,
-                    keys,
-                    address,
-                    who=who,
-                    serializers=serializers,
-                    max_connections=False,
+                retry_operation(
+                    partial(
+                        get_data_from_worker,
+                        rpc,
+                        keys,
+                        address,
+                        who=who,
+                        serializers=serializers,
+                        max_connections=False,
+                    ),
+                    operation="get_data_from_worker",
                 ),
                 name=f"get-data-from-{address}",
             )
             for address, keys in d.items()
         }
-        response = {}
+        response: dict[str, object] = {}
         for worker, c in coroutines.items():
             try:
                 r = await c
@@ -90,8 +102,11 @@ async def gather_from_workers(who_has, rpc, close=True, serializers=None, who=No
         bad_addresses |= {v for k, v in rev.items() if k not in response}
         results.update(response)
 
-    bad_keys = {k: list(original_who_has[k]) for k in all_bad_keys}
-    return (results, bad_keys, list(missing_workers))
+    return (
+        results,
+        {k: list(original_who_has[k]) for k in all_bad_keys},
+        list(missing_workers),
+    )
 
 
 class WrappedKey:
@@ -115,7 +130,7 @@ class WrappedKey:
 _round_robin_counter = [0]
 
 
-async def scatter_to_workers(nthreads, data, rpc=rpc, report=True):
+async def scatter_to_workers(nthreads, data, rpc=rpc):
     """Scatter data directly to workers
 
     This distributes data in a round-robin fashion to a set of workers based on
@@ -139,15 +154,7 @@ async def scatter_to_workers(nthreads, data, rpc=rpc, report=True):
 
     rpcs = {addr: rpc(addr) for addr in d}
     try:
-        out = await All(
-            [
-                rpcs[address].update_data(
-                    data=v,
-                    report=report,
-                )
-                for address, v in d.items()
-            ]
-        )
+        out = await All([rpcs[address].update_data(data=v) for address, v in d.items()])
     finally:
         for r in rpcs.values():
             await r.close_rpc()
@@ -162,7 +169,89 @@ async def scatter_to_workers(nthreads, data, rpc=rpc, report=True):
 collection_types = (tuple, list, set, frozenset)
 
 
-def unpack_remotedata(o, byte_keys=False, myset=None):
+def _namedtuple_packing(o: Any, handler: Callable[..., Any]) -> Any:
+    """Special pack/unpack dispatcher for namedtuples respecting their potential constructors."""
+    assert is_namedtuple_instance(o)
+    typ = type(o)
+    if hasattr(o, "__getnewargs_ex__"):
+        args, kwargs = o.__getnewargs_ex__()
+        handled_args = [handler(item) for item in args]
+        handled_kwargs = {k: handler(v) for k, v in kwargs.items()}
+        return typ(*handled_args, **handled_kwargs)
+    args = o.__getnewargs__() if hasattr(typ, "__getnewargs__") else o
+    handled_args = [handler(item) for item in args]
+    return typ(*handled_args)
+
+
+def _unpack_remotedata_inner(
+    o: Any, byte_keys: bool, found_keys: set[WrappedKey]
+) -> Any:
+    """Inner implementation of `unpack_remotedata` that adds found wrapped keys to `found_keys`"""
+
+    typ = type(o)
+    if typ is tuple:
+        if not o:
+            return o
+        if type(o[0]) is SubgraphCallable:
+            # Unpack futures within the arguments of the subgraph callable
+            futures: set[WrappedKey] = set()
+            args = tuple(_unpack_remotedata_inner(i, byte_keys, futures) for i in o[1:])
+            found_keys.update(futures)
+
+            # Unpack futures within the subgraph callable itself
+            sc: SubgraphCallable = o[0]
+            futures = set()
+            dsk = {
+                k: _unpack_remotedata_inner(v, byte_keys, futures)
+                for k, v in sc.dsk.items()
+            }
+            future_keys: tuple = ()
+            if futures:  # If no futures is in the subgraph, we just use `sc` as-is
+                found_keys.update(futures)
+                future_keys = (
+                    tuple(stringify(f.key) for f in futures)
+                    if byte_keys
+                    else tuple(f.key for f in futures)
+                )
+                inkeys = tuple(sc.inkeys) + future_keys
+                sc = SubgraphCallable(dsk, sc.outkey, inkeys, sc.name)
+            return (sc,) + args + future_keys
+        else:
+            return tuple(
+                _unpack_remotedata_inner(item, byte_keys, found_keys) for item in o
+            )
+    elif is_namedtuple_instance(o):
+        return _namedtuple_packing(
+            o,
+            partial(
+                _unpack_remotedata_inner, byte_keys=byte_keys, found_keys=found_keys
+            ),
+        )
+
+    if typ in collection_types:
+        if not o:
+            return o
+        outs = [_unpack_remotedata_inner(item, byte_keys, found_keys) for item in o]
+        return typ(outs)
+    elif typ is dict:
+        if o:
+            return {
+                k: _unpack_remotedata_inner(v, byte_keys, found_keys)
+                for k, v in o.items()
+            }
+        else:
+            return o
+    elif issubclass(typ, WrappedKey):  # TODO use type is Future
+        k = o.key
+        if byte_keys:
+            k = stringify(k)
+        found_keys.add(o)
+        return k
+    else:
+        return o
+
+
+def unpack_remotedata(o: Any, byte_keys: bool = False) -> tuple[Any, set]:
     """Unpack WrappedKey objects from collection
 
     Returns original collection and set of all found WrappedKey objects
@@ -189,61 +278,8 @@ def unpack_remotedata(o, byte_keys=False, myset=None):
     >>> unpack_remotedata(rd, byte_keys=True)
     ("('x', 1)", {WrappedKey('('x', 1)')})
     """
-    if myset is None:
-        myset = set()
-        out = unpack_remotedata(o, byte_keys, myset)
-        return out, myset
-
-    typ = type(o)
-
-    if typ is tuple:
-        if not o:
-            return o
-        if type(o[0]) is SubgraphCallable:
-            sc = o[0]
-            futures = set()
-            dsk = {
-                k: unpack_remotedata(v, byte_keys, futures) for k, v in sc.dsk.items()
-            }
-            args = tuple(unpack_remotedata(i, byte_keys, futures) for i in o[1:])
-            if futures:
-                myset.update(futures)
-                futures = (
-                    tuple(stringify(f.key) for f in futures)
-                    if byte_keys
-                    else tuple(f.key for f in futures)
-                )
-                inkeys = sc.inkeys + futures
-                return (
-                    (SubgraphCallable(dsk, sc.outkey, inkeys, sc.name),)
-                    + args
-                    + futures
-                )
-            else:
-                return o
-        else:
-            return tuple(unpack_remotedata(item, byte_keys, myset) for item in o)
-    elif is_namedtuple_instance(o):
-        return typ(*[unpack_remotedata(item, byte_keys, myset) for item in o])
-
-    if typ in collection_types:
-        if not o:
-            return o
-        outs = [unpack_remotedata(item, byte_keys, myset) for item in o]
-        return typ(outs)
-    elif typ is dict:
-        if o:
-            return {k: unpack_remotedata(v, byte_keys, myset) for k, v in o.items()}
-        else:
-            return o
-    elif issubclass(typ, WrappedKey):  # TODO use type is Future
-        k = o.key
-        if byte_keys:
-            k = stringify(k)
-        myset.add(o)
-        return k
-    else:
-        return o
+    found_keys: set[Any] = set()
+    return _unpack_remotedata_inner(o, byte_keys, found_keys), found_keys
 
 
 def pack_data(o, d, key_types=object):
@@ -277,6 +313,8 @@ def pack_data(o, d, key_types=object):
         return typ([pack_data(x, d, key_types=key_types) for x in o])
     elif typ is dict:
         return {k: pack_data(v, d, key_types=key_types) for k, v in o.items()}
+    elif is_namedtuple_instance(o):
+        return _namedtuple_packing(o, partial(pack_data, d=d, key_types=key_types))
     else:
         return o
 
@@ -313,15 +351,19 @@ def subs_multiple(o, d):
             return o
 
 
+T = TypeVar("T")
+
+
 async def retry(
-    coro,
-    count,
-    delay_min,
-    delay_max,
-    jitter_fraction=0.1,
-    retry_on_exceptions=(EnvironmentError, IOError),
-    operation=None,
-):
+    coro: Callable[[], Coroutine[Any, Any, T]],
+    count: int,
+    delay_min: float,
+    delay_max: float,
+    jitter_fraction: float = 0.1,
+    retry_on_exceptions: type[BaseException]
+    | tuple[type[BaseException], ...] = (EnvironmentError, IOError),
+    operation: str | None = None,
+) -> T:
     """
     Return the result of ``await coro()``, re-trying in case of exceptions
 
@@ -371,7 +413,13 @@ async def retry(
     return await coro()
 
 
-async def retry_operation(coro, *args, operation=None, **kwargs):
+# ParamSpec is not supported here due to the additional "operation" kwarg
+async def retry_operation(
+    coro: Callable[..., Coroutine[Any, Any, T]],
+    *args: object,
+    operation: str | None = None,
+    **kwargs: object,
+) -> T:
     """
     Retry an operation using the configuration values for the retry parameters
     """
